"""
Python script by Tobias Polzin, polzin@gmx.de
All rights reserved.

<title>
evol.py, Version 24.3.00
</title>
<long>
Evolutionsstrategies: Powerful global optimisation<br>
Basic class for a global optimisation strategie called
<a href=http://ls11-www.informatik.uni-dortmund.de/people/kursawe/Demos/EvoNet/EStheory.html>'Evolutionsstrategie' by Prof.Schwefel</a>.  <br>
Parameters:
<tt>(n,action,startx=0,startd=0,pop=20,child=10,life=5,gen=100,matestrat=0,startdd=0.01)</tt>
<dl>
<dt><tt>n</tt><dd>dimesion of problem (number of float parameters for action)
<dt><tt>action</tt><dd>the function <tt>action</tt> gets <tt>n</tt>
parameters and gives the function value
<dt><tt>startx</tt><dd>startvalues for search, must be a
<tt>n</tt>-item list or 0, (default=0)
<dt><tt>startd</tt><dd>startvalues for stepsize (default=0)
<dt><tt>pop</tt><dd>maximal size of population (worst species are killed) (default=20)
<dt><tt>child</tt><dd>childs per generation(default=10)
<dt><tt>life</tt><dd>lifetime of each species(default=5)
<dt><tt>gen</tt><dd>number of generations(default=100)
<dt><tt>matestrat</tt><dd>recombination strategie: 0=discrete,1=itermediate (default=0)
<dt><tt>startdd</tt><dd>stepsize stepsize (default=0.01)
</dl>
<br>
Examples:
<pre>
def test(x,y,z):
    return (x-2)*(x-2)+y*y+(z+1)*(z+1)

e=evol(3,test,matestrat=0)
</pre>
</long>
"""
import math
import random
import whrandom
import local

def convtoint(f): return math.floor(f)
def convtonear1(f): return 1-1.0/f
def convtorange(f,lo,up): return max(lo,min(up,f))
def convtopos(f): return max(0,f)

def listtotuple(l):
    t=()
    for i in l: t=t+(i,)
    return t

class item:
    def __init__(self,x,d,dd):
        self.x=x
        self.d=d
        self.dd=dd
    def set(self,action,birth):
        self.value=apply(action,listtotuple(self.x))
        self.birth=birth
        return self
        
def matei(i,j,s):
    return item(matel(i.x,j.x,s),
                matel(i.d,j.d,s),
                mates(i.dd,j.dd,s))
def matel(i,j,s):
    return map(lambda i,j,s=s:mates(i,j,s),i,j)
def mates(i,j,s):
    if s:
        return i+j/2
    else:
        return whrandom.choice([i,j])
    
class evol:
    def makeitem(self,x,d,dd):
        return item(x,d,dd).set(self.action,self.iter)
    def mutate(self,i):
        dd=i.dd * random.normalvariate(1,0.1)
        alpha=random.normalvariate(0,dd)
        d=map(lambda d,dd=i.dd,alpha=alpha:
                       d * math.exp(alpha + random.normalvariate(0,dd)),
                       i.d)
        x=map(lambda x,dx:
                      random.normalvariate(x,dx),
                      i.x,
                      i.d)
        i=self.makeitem(x,d,dd)
        if i.value<self.best.value:
            print "  improvement: ",self.best.value,i.value
            self.best=i
        return i
                    
    def __init__(self,n,action,startx=0,startd=0,pop=20,child=10,
                 life=5,gen=100,matestrat=0,startdd=0.01):
        if not startx:
            startx=n*[0.]
        if not startd:
            startd=map(lambda a:a/10+0.2,startx)
        self.iter=0
        self.action=action
        self.best=self.makeitem(startx,startd,startdd)
        items=[]
        for i in range(pop):
            items.append(self.mutate(self.best))

        for self.iter in range(gen):
            for i in range(child):
                items.append(self.mutate(matei
                                         (whrandom.choice(items),
                                          whrandom.choice(items),
                                          matestrat)))

            killtime=self.iter-life
            items=filter(lambda i,killtime=killtime:i.birth>killtime,
                         items)
            if len(items)>pop:
                items=local.sortfor(lambda i:i.value,items)
                del items[pop:]
            print "   -", self.iter,self.best.value,items[0].value,len(items)
        self.items=items
        print self.best.x

        
        
                                                       
                       





